/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/** 
 * @file applications/afs/encoding/requestmanager.c
 * @brief The RequestManager keeps track of and re-issues queries
 * @author Christian Grothoff
 **/ 

#include "config.h"

#include <errno.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "util/checksum.h"
#include "util/logging.h"
#include "util/tcp_return.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "applications/afs/util/policy.h"
#include "applications/afs/encoding/requestmanager.h"
#include "applications/afs/protocol/messages.h"
#include "util/port.h"

#define DEBUG_REQUESTMANAGER NO

#define MIN_CONNECTED 10

static unsigned int connectedPeers = MIN_CONNECTED;

static void queryPeerCount(GNUNET_TCP_SOCKET * sock) {
  CS_HEADER q;

  q.size = htons(sizeof(CS_HEADER));
  q.tcpType = htons(CS_PROTO_CLIENT_COUNT);
  if (OK == writeToSocket(sock, &q)) {
    if (SYSERR == readTCPResult(sock, &connectedPeers))
      LOG(LOG_WARNING,
	  "WARNING: gnunetd failed to respond to query for number of connected peers\n");
  } else
    LOG(LOG_WARNING,
	"WARNING: could not send message to gnuentd via TCP\n");
  if (connectedPeers < MIN_CONNECTED)
    connectedPeers = MIN_CONNECTED;
}

/**
 * how often do we try a query  (in a row, it will be
 * re-scheduled later!) 
 **/
#define MAX_TRIES 50

/**
 * Print the contents of the request manager. For debugging.
 **/
void printRequestManager(RequestManager * this) {
  int i;
  HexName hex;

  MUTEX_LOCK(&this->lock);
  LOG(LOG_DEBUG,
      "DEBUG: RM TTL %u duplicates %d\n",
      this->initialTTL, 
      this->duplicationEstimate);
  for (i=0;i<this->requestListIndex;i++) {
    IFLOG(LOG_DEBUG,
	  hash2hex(&this->requestList[i]->message->queries[0], 
		   &hex));
    LOG(LOG_DEBUG,
	"DEBUG: %4i: %s for node %d (%d tries)\n",
	i, 
	&hex, 
	this->requestList[i]->receiverNode,
	this->requestList[i]->tries);
  }  
  MUTEX_UNLOCK(&this->lock);
}

/**
 * Send the request from the requestList[requestIndex] out onto
 * the network. If the request has been pending for too long, put it
 * back into the requestList and move another one in its place
 * instead.
 *
 * @param this the RequestManager
 * @param requestIndex the index of the Request to issue
 **/
static void issueRequest(RequestManager * this,
			 int requestIndex) {
  NodeContext * nc;
  RequestEntry * entry;
  cron_t now;
  int ttl;
  HexName hex;
  CS_HEADER * msg;
  GNUNET_TCP_SOCKET * sock;
  int prevttl;
  int prevpri;
  cron_t prevlt;
  int ok;

  cronTime(&now);
  entry = this->requestList[requestIndex];
  if (entry->lasttime == 0) {
    ttl = this->initialTTL;
    prevttl = ttl;
  } else {
    ttl = ntohl(entry->message->ttl);
    prevttl = ttl;
    ttl += randomi(4*TTL_DECREMENT);
    if (ttl > 4 * TTL_DECREMENT * (int) log(connectedPeers))
      ttl = 4 * TTL_DECREMENT * ((int)log(connectedPeers)) + 
	randomi(2*TTL_DECREMENT);
    if (ttl > 5 * this->initialTTL) {
      this->initialTTL = (this->initialTTL * 99 + ttl) / 100; /* grow initialTTL a bit */
      ttl = 5 * this->initialTTL + randomi(100); /* and in turn bound the ttl */
    }
  }
  prevlt = entry->lasttime;
  entry->lasttime = now;
  if (NO == checkAnonymityPolicy(AFS_CS_PROTO_QUERY,
				 ntohs(entry->message->header.size)+sizeof(HostIdentity))) {
#if DEBUG_REQUESTMANAGER
    LOG(LOG_DEBUG,
	"DEBUG: not sending query %s due to anonymity policy!\n",
	&hex);
#endif
    return; 
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&entry->message->queries[0],
		 &hex));
  entry->message->ttl = htonl(ttl);
  prevpri = ntohl(entry->message->priority);
  if (prevpri > connectedPeers * 1000)
    prevpri = connectedPeers * 1000; /* bound! */
  entry->tries++;
  entry->message->priority 
    = htonl(prevpri+
	    randomi(entry->tries + prevpri));
#if DEBUG_REQUESTMANAGER
  LOG(LOG_DEBUG,
      "DEBUG: sending query %s with ttl %d and priority %u (%d)\n",
      &hex,
      ttl,
      ntohl(entry->message->priority),
      this->initialTTL);
#endif
  nc = (NodeContext *)entry->data;
  nc->stats.requestsPending 
    = this->requestListIndex;
  nc->stats.requestsSent 
    = this->requestListIndex;
  nc->stats.currentTTL 
    = ttl;
  nc->stats.duplicationEstimate 
    = this->duplicationEstimate;
  nc->pmodel(&nc->stats, nc->data);    

  if (0 == (entry->tries % (MAX_TRIES * 50))) {
    IFLOG(LOG_WARNING,
	  hash2hex(&entry->message->queries[0],
		   &hex));
    LOG(LOG_WARNING,
	"WARNING: %s seems to be not available on the network\n",
	&hex);
    entry->receiverNode->print(entry->receiverNode, 0);
  }
  msg = MALLOC(ntohs(entry->message->header.size));
  memcpy(msg,
	 entry->message,
	 ntohs(entry->message->header.size));
  sock = this->sock;

  ok = SYSERR;
  if (sock != NULL) { 
    /* destroyRM may set this->sock to NULL at ANY point! */
    ok = writeToSocketNonBlocking(sock, 
				  msg);
    if (ok == SYSERR) 
      LOG(LOG_WARNING,
	  "WARNING: could not send request to gnunetd, "
	  "is it running?\n");    
  }
  if (ok != OK) {
    /* we did not send this entry, revert! */
    entry->message->ttl = htonl(prevttl);
    entry->message->priority = htonl(prevpri);
    entry->lasttime = prevlt;
    entry->tries--;
  } else {
    if (entry->tries > 1) {
      static int ptries = 2;
      if (entry->tries > ptries) { /* we're a bit more lax about retransmission
				      than TCP -- and each time we start with
				      congestion control, we make it harder for
				      the next round! */
	ptries++;
	/* we performed retransmission, treat as congestion (RFC 2001) */
	this->ssthresh 
	  = this->congestionWindow / 2;
	if (this->ssthresh < 2)
	  this->ssthresh = 2;
	this->congestionWindow 
	  = 1; /* where RSS is 1 */    
      }
      nc->stats.totalRetries++;
      nc->stats.currentRetries++;      
    }
  }
  FREE(msg);
}

/**
 * Cron job that re-issues requests. Should compute how long to sleep
 * (min ttl until next job is ready) and re-schedule itself
 * accordingly!
 **/
static void requestJob(RequestManager * this) {
  cron_t minSleep;
  cron_t now;
  cron_t delta;
  int i;  
  int pending;

#if DEBUG_REQUESTMANAGER
  LOG(LOG_CRON,
      "CRON: requestJob %x running\n",
      (int)this);
#endif
  MUTEX_LOCK(&this->lock);
  cronTime(&now);
  pending = 0;

  for (i=0;i<this->requestListIndex;i++) {
    if (this->requestList[i]->lasttime +
	ntohl(this->requestList[i]->message->ttl) >= now) 
      pending++;
  }

  minSleep = 5 * cronSECONDS; /* max-sleep! */
  for (i=0;i<this->requestListIndex;i++) {
    if ((this->requestList[i]->lasttime +
	 ntohl(this->requestList[i]->message->ttl)) <= now + TTL_DECREMENT) {
      if ( (pending < this->congestionWindow) ||
	   (randomi(this->requestListIndex) == 0 /* avoid no-start */) ) {
	delta = ntohl(this->requestList[i]->message->ttl) + 10 * cronMILLIS;    
	issueRequest(this, i);
	pending++;
      } else {
	LOG(LOG_DEBUG,
	    "DEBUG: congestion control: %d pending, %d window\n",
	    pending,
	    this->congestionWindow);
	delta = 0;
      }
    } else {
      delta = (this->requestList[i]->lasttime +
	       ntohl(this->requestList[i]->message->ttl)) - now;
      LOG(LOG_DEBUG,
	  "DEBUG: request %d (%u) is still pending for %us\n",
	  i,
	  ntohl(this->requestList[i]->message->ttl),
	  (unsigned int) (delta / cronSECONDS));	
    }
      
    if ( delta < minSleep )
      minSleep = delta;
  }
  if (minSleep < cronMILLIS * 10)
    minSleep = cronMILLIS * 10; /* maximum resolution: 10ms */
  if (this->requestListIndex > 0) {
#if DEBUG_REQUESTMANAGER
    LOG(LOG_CRON,
	"CRON: scheduling next run for in %dms\n",
	minSleep);
#endif
    addCronJob((CronJob)&requestJob, 
	       minSleep, 
	       0, 
	       this);
  }
#if DEBUG_REQUESTMANAGER
  else
    LOG(LOG_DEBUG,
	"DEBUG: no more jobs pending, cron not renewed!\n");
#endif
  MUTEX_UNLOCK(&this->lock);
}

/**
 * Listen on socket and receive messages. Call requestManagerReceive
 * on every reply. Never returns, if the RM dies, it will cancel us.
 **/
static void * receiveThread(RequestManager * this) {
  CS_HEADER * buffer;
  GNUNET_TCP_SOCKET * sock;

  while (this->sock != NULL) {
    MUTEX_LOCK(&this->lock);
    sock = this->sock;
    MUTEX_UNLOCK(&this->lock);
    if (sock == NULL)
      break;
    buffer = NULL;
    if (SYSERR == readFromSocket(sock,
				 &buffer) ) {
      if (this->sock == NULL)
	break;
      LOG(LOG_WARNING,
	  "WARNING: requestmanager-receive thread could not "
	  "read data from gnunetd, is the server running?\n");
      sleep(15);
      continue;
    }
    if ( (ntohs(buffer->tcpType) == AFS_CS_PROTO_RESULT_CHK) &&
	 (ntohs(buffer->size) == sizeof(AFS_CS_RESULT_CHK) ) ) {
      requestManagerReceive(this, 
			    (AFS_CS_RESULT_CHK*) buffer);
    } else {
      /* This should no longer happen, but better check than sorry... */
      LOG(LOG_WARNING,
	  "WARNING: received unexpected message (%d) from gnunetd. "
	  "(this is a bug, though we can probably recover gracefully)\n",
	  ntohs(buffer->tcpType));
      MUTEX_LOCK(&this->lock);
      releaseClientSocket(this->sock);
      this->sock = getClientSocket();
      MUTEX_UNLOCK(&this->lock);
    }
    FREE(buffer);
  }
  SEMAPHORE_UP(this->destroySignal);
  return NULL;
}

/**
 * Create a request manager. Will create the request manager
 * datastructures and also connect to gnunetd. Creates thread that
 * listens to gnunetd replies and another thread that periodically
 * re-issues the queries. Use destroyRequestManager to abort and/or to
 * free resources after the download is complete. The callback method
 * in nc will be invoked to notify the caller of the download progress
 * such that it is possible to tell when we are done.
 *
 * @return NULL on error
 **/
RequestManager * createRequestManager() {
  RequestManager * rm;

  rm = MALLOC(sizeof(RequestManager));
  MUTEX_CREATE_RECURSIVE(&rm->lock);
  rm->requestListIndex 
    = 0;
  rm->requestListSize  
    = 0;
  rm->requestList      
    = NULL;
  GROW(rm->requestList,
       rm->requestListSize,
       256);
  rm->initialTTL
    = 5 * cronSECONDS;
  /* RFC 2001 suggests to use 1 segment size initially;
     Given 1500 octets per message in GNUnet, we would
     have 2-3 queries of maximum size (552); but since
     we are multi-casting to many peers at the same
     time AND since queries can be much smaller,
     we multiply 3 by the number of connected peers+1 */
  rm->congestionWindow
    = 1; /* RSS is 1 */
  rm->ssthresh
    = 65535;
  rm->duplicationEstimate
    = 0;
  rm->sock
    = getClientSocket();
  rm->destroySignal
    = SEMAPHORE_NEW(0);
  if (rm->sock == NULL) {
    LOG(LOG_WARNING,
	"WARNING: could not create socket to connect to gnunetd\n");
    SEMAPHORE_UP(rm->destroySignal);
    destroyRequestManager(rm);
    return NULL;
  }
  memset(&rm->receiveThread_, 0, sizeof(pthread_t));
  queryPeerCount(rm->sock);
  if (0 != PTHREAD_CREATE(&rm->receiveThread_,
			  (void*(*)(void*))&receiveThread,
			  rm, 
			  256*1024)) {
    LOG(LOG_ERROR,
	    "ERROR: could not create receiveThread (%s)\n",
	  strerror(errno));
    SEMAPHORE_UP(rm->destroySignal);
    destroyRequestManager(rm);
    return NULL;
  }
  return rm;
}

/**
 * Destroy the resources associated with a request manager.
 * Invoke this method to abort the download or to clean up
 * after the download is complete.
 *
 * @param this the request manager struct from createRequestManager
 **/
void destroyRequestManager(RequestManager * this) {
  GNUNET_TCP_SOCKET * sock;
  int i;
  void * unused;

  suspendCron();
  MUTEX_LOCK(&this->lock);
  sock = this->sock;
  this->sock = NULL;
  MUTEX_UNLOCK(&this->lock);
  closeSocketTemporarily(sock); /* unblock RM thread */
  SEMAPHORE_DOWN(this->destroySignal);
  SEMAPHORE_FREE(this->destroySignal);
  PTHREAD_JOIN(this->receiveThread_, &unused);
  destroySocket(sock);
  delCronJob((CronJob)&requestJob, 
	     0, 
	     this);
  MUTEX_DESTROY(&this->lock);
  for (i=0;i<this->requestListIndex;i++) 
    FREENONNULL(this->requestList[i]);
  GROW(this->requestList,
       this->requestListSize,
       0);
  FREE(this);
  releaseClientSocket(sock);
  resumeCron();
}

/**
 * This method receives data corresponding to the indicated filename
 * (hashcode). Finds the Listener that scheduled this request and drop
 * it from the list of pending requests.<p>
 * 
 * @param this the request manager struct from createRequestManager
 * @param msg the message received from gnunetd
 **/ 
void requestManagerReceive(RequestManager * this,
			   AFS_CS_RESULT_CHK * msg) {
  int pos;
  int i;
  int j;
  HashCode160 query;
  RequestEntry * entry;

  /* check type of reply msg, fill in query */
  hash(&msg->result,
       sizeof(CONTENT_Block),
       &query);  
  suspendCron();
  MUTEX_LOCK(&this->lock);
  pos = -1;  
  /* find which query matches the reply, call the callback
     and recycle the slot */
  for (i=0;i<this->requestListIndex;i++) {
    AFS_CS_QUERY * acq;

    acq = this->requestList[i]->message;
    j = (ntohs(acq->header.size)-sizeof(AFS_CS_QUERY))/sizeof(HashCode160);    
    while ( j > 0 ) {
      j--;
      if (equalsHashCode160(&query, 
			    &acq->queries[j])) 
	pos = i;
    }
  }
  if (pos == -1) {
    this->duplicationEstimate++;

    /* duplicate reply, treat as congestion (RFC 2001) */
    if (this->duplicationEstimate % 3 == 0) {
      this->ssthresh = this->congestionWindow / 2;
      if (this->ssthresh < 2)
	this->ssthresh = 2;
      this->congestionWindow 
	= this->ssthresh + 1;
    } else {
      this->congestionWindow += 1;
    }

#if DEBUG_REQUESTMANAGER
    LOG(LOG_DEBUG, 
	"DEBUG: RequestManager: received useless data (%d, %ds)!\n",
	this->duplicationEstimate,
	this->initialTTL / cronSECONDS);   
#endif
    MUTEX_UNLOCK(&this->lock);
    resumeCron();
    return;
  } 
  entry = this->requestList[pos];
  
  if ( (entry->lasttime < cronTime(NULL)) &&
       (entry->lasttime != 0) ) /* we can get replies for things that we never
				   asked for, do not count the ttl for those!
				   Also, some ttls are adjusted into the future, they
				   should also not count! */ {
    this->initialTTL = ((this->initialTTL) + ntohl(entry->message->ttl) - 100) / 2;

    /* RFC 2001: increase cwnd; note that we can't really discriminate between
       slow-start and cong. control mode since our RSS is too small... */
    if (this->congestionWindow < this->ssthresh)
      this->congestionWindow += 2; /* slow start */
    else
      this->congestionWindow += 1; /* slower start :-) */
  }
  /* and finally use the entry to notify the node
     that we got a reply! */
#if DEBUG_REQUESTMANAGER
  LOG(LOG_DEBUG,
      "DEBUG: request manager receives data for %x\n",
      entry->receiverNode);
#endif

  if (SYSERR == entry->receiver(entry->receiverNode,
				&query,
				msg,
				this,
				entry->data)) {
    int i;
    
    /* ABORT download, receiver method has
       already notified the controller via
       the pmodel callback, we need to stop
       requesting... */
#if DEBUG_REQUESTMANAGER
    LOG(LOG_DEBUG,
	"DEBUG: entry->receiver aborted download!\n");
#endif
    for (i=0;i<this->requestListIndex;i++)
      FREENONNULL(this->requestList[i]);
    this->requestListIndex = 0;
    delCronJob((CronJob)&requestJob,
	       0, 
	       this);   
  }
  MUTEX_UNLOCK(&this->lock);
  resumeCron();
}

/**
 * Queue a request for execution.
 * 
 * @param this the request manager struct from createRequestManager
 * @param node the node to call once a reply is received
 * @param callback the method to invoke
 * @param data the data argument to the Listener
 * @param message the query to send to gnunetd, freed by callee!
 **/
void requestManagerRequest(RequestManager * this,
			   Block * node,
			   Listener callback,
			   void * data,
			   AFS_CS_QUERY * message) {
  RequestEntry * entry;

#if DEBUG_REQUESTMANAGER
  LOG(LOG_DEBUG,
      "DEBUG: requestManagerRequest for %x with callback %x\n",
      node,
      callback);
#endif
  entry = MALLOC(sizeof(RequestEntry));
  entry->message 
    = message;
  entry->lasttime 
    = 0; /* never sent */
  entry->receiver 
    = callback;
  entry->receiverNode
    = node;
  entry->data 
    = data;
  entry->tries 
    = 0; /* not tried so far */

  suspendCron();
  MUTEX_LOCK(&this->lock);   
  /* can we add to current list & issue instantly? */
  if (this->requestListSize == this->requestListIndex) 
    GROW(this->requestList,
	 this->requestListSize,
	 this->requestListSize*2);
  this->requestList[this->requestListIndex++] = entry;
  delCronJob((CronJob)&requestJob,
	     0, 
	     this);
#if DEBUG_REQUESTMANAGER
  LOG(LOG_CRON,
      "CRON: scheduling next run for now!\n");
#endif
  addCronJob((CronJob)&requestJob,
	     0, 
	     0,
	     this); 
  MUTEX_UNLOCK(&this->lock);
  resumeCron();
  return;
}

/**
 * Assert that there are no pending requests for this node.
 **/
void requestManagerAssertDead(RequestManager * this,
			      Block * node) {
  int i;

  if (this == NULL)
    return; /* do not check */
  MUTEX_LOCK(&this->lock);
  for (i=0;i<this->requestListIndex;i++)
    if (this->requestList[i]->receiverNode == node) 
      errexit("FATAL: node %x is being destroyed while request is pending\n",
	      node);
  MUTEX_UNLOCK(&this->lock);
}

/**
 * Update a request.  This method is used to selectively change a
 * query or drop it entirely.
 *
 * @param this the request manager struct from createRequestManager
 * @param node the block for which the request is updated
 * @param msg the new query message for that node, NULL for 
 *        none (then the request is dropped)
 **/
void requestManagerUpdate(RequestManager * this,
			  Block * node,
			  AFS_CS_QUERY * msg) {
  int i;

#if DEBUG_REQUESTMANAGER
  LOG(LOG_DEBUG,
      "DEBUG: updating request for %x to %x\n",
      node, 
      msg);
#endif
  MUTEX_LOCK(&this->lock);
  for (i=0;i<this->requestListIndex;i++) {
    if (this->requestList[i]->receiverNode == node) {
      if (msg != NULL) { /* update */
	msg->ttl 
	  = htonl(ntohl(this->requestList[i]->message->ttl)/2);
	msg->priority 
	  = htonl(ntohl(this->requestList[i]->message->priority)/2);
	FREE(this->requestList[i]->message);
	this->requestList[i]->message = msg;	
	/* also wait a bit longer before re-issueing the
	   request, after all, we got at least one of the
	   replies! */
	this->requestList[i]->lasttime 
	  = cronTime(NULL) + 2*TTL_DECREMENT; /* add 2*TTL grace time if we got a reply
						 to a multi-query; this dramatically
						 reduces the amount of "useless"
						 (duplicate) replies we get! */
      } else { /* delete */
        /* update stats */
        if(this->requestList[i]->tries > 1)
          ((NodeContext *)(this->requestList[i]->data))->stats.currentRetries 
	     -= (this->requestList[i]->tries - 1);
	FREE(this->requestList[i]->message);
	FREE(this->requestList[i]);
	this->requestList[i] 
	  = this->requestList[--this->requestListIndex];
	this->requestList[this->requestListIndex] 
	  = NULL;	
      }
      MUTEX_UNLOCK(&this->lock);
      return; /* found! */
    }
  }
  MUTEX_UNLOCK(&this->lock);  
}

/* end of requestmanager.c */
