/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/main.c
 * @brief This is the main file for the gtk+ user interface.
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 *
 * Basic structure of the code is this:
 * main
 * -> search -> saveas -> download
 * -> insert
 * -> about
 *
 * TODO: 
 * 
 * - Anticipating future off-band sharing of gnunet keys,
 *   add possibility to download files by specifying their keys
 *   (hash1, hash2, crc, size) manually (we may want to standardize
 *   the URI format for GNUnet keys a bit more, first)
 **/

#include "config.h"

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "util/getopt.h"
#include "util/initialize.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/storage.h"
#include "util/xmalloc.h"
#include "applications/afs/afs.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/search.h"
#include "applications/afs/gtkui/insert.h"
#include "applications/afs/gtkui/about.h"
#include "applications/afs/util/policy.h"
#include "applications/common/printhelp.h"
#include "util/configuration.h"

/**
 * This semaphore can be used to prevent the main window
 * from killing GTK at an unhealty time. It is used by
 * the un-interruptable but GUI-updating insert thread.
 **/
Semaphore * refuseToDie;

static GtkWidget * main_window_input_line = NULL;
static GtkNotebook * notebook = NULL;

/**
 * Remove the active page from the search results notebook.
 * The respective search will be stopped as well
 * (by a callback assigned to the page earlier on).
 **/
static void close_search(GtkButton * button,
			 gpointer dummy) {
  gint page;
  
  page = gtk_notebook_get_current_page(notebook);
  gtk_notebook_remove_page(notebook, page);
  /* Need to refresh the widget --
     This forces the widget to redraw itself. */
  gtk_widget_draw(GTK_WIDGET(notebook), NULL);
}

/**
 * This method is called whenever the user clicks the
 * search button of the main window. 
 * 
 * @param widget not used
 * @param notebook the notebook where the opened result window will be put.
 **/
static void search(GtkWidget * widget,
		   GtkNotebook * notebook) {
  gchar * searchString;
  gchar * searchStringStart;
  GtkWidget * label;
  GtkWidget * tmp;
  int i; 

  searchString 
    = gtk_entry_get_text(GTK_ENTRY(main_window_input_line));

  if (searchString == NULL) {
    guiMessage("searchString == NULL\n");
    return;
  }

  searchString = STRDUP(searchString);

  /* Remove heading spaces (parsing bugs if not) */  
  i = strlen(searchString)-1;
  while ( (searchString[i]==' ') && 
	  (i>=0) ) {
    searchString[i] = '\0';
    i--;
  }

  searchStringStart = searchString;
  while (*searchStringStart == ' ')
    searchStringStart++;

  if (*searchStringStart=='\0') {
    guiMessage("No search key given!\n");
    FREE(searchString);
    return;
  }

  /* add a new page in the notebook with the
     search results. getSearchWindow returns
     the page */
  tmp = getSearchWindow(searchStringStart);
  if (tmp != NULL) {
    /* add a new notebook for the search results */
    label = gtk_label_new(searchStringStart);
    gtk_notebook_append_page(notebook, tmp, label);
    gtk_widget_show(tmp);
  }

  /* reset search line to empty */
  gtk_entry_set_text(GTK_ENTRY(main_window_input_line), 
		     ""); 
  FREE(searchString);
}

/**
 * Exit the application (called when the main window
 * is closed or the user selects File-Quit).
 **/
static void destroyMain(GtkWidget * widget,
			gpointer data) {
  stopCron();
  if (notebook != NULL) {
    int i = 0;    
    while(gtk_notebook_get_nth_page(notebook, 0) != NULL) {
      LOG(LOG_DEBUG, 
	  "DEBUG: Removing search page %d\n", 
	  i++);
      gtk_notebook_remove_page(notebook, 0);
    }
  }
  startCron();
  gdk_threads_leave(); /* avoid deadlock! */
  SEMAPHORE_DOWN(refuseToDie);
  gdk_threads_enter();
  gtk_main_quit(); /* back to main method! */
}

/**
 * Method called from menu bar "File-Quit". Wrapper around
 * destroy.
 **/ 
static void destroy_stub(void) {
  destroyMain(NULL, NULL);
}

/**
 * The pulldown menus.
 **/
static GtkItemFactoryEntry menu_items[] = {
  { "/_File",         
    NULL, 
    NULL,
    0,
    "<Branch>" },
  { "/File/_Insert",  
    "<control>I", 
    openSelectFile, 
    0, 
    NULL },
  { "/File/sep1",     
    NULL, 
    NULL, 
    0,
    "<Separator>" },
  { "/File/_Abort search",      
    "<control>A", 
    close_search, 
    0, 
    NULL },
  { "/File/sep1",     
    NULL,        
    NULL, 
    0, 
    "<Separator>" },
  { "/File/_Quit",     
    "<control>Q", 
    destroy_stub,
    0, 
    NULL },
/*
  { "/_Options",     
    NULL,  
    NULL, 
    0,
    "<Branch>" },
  { "/Options/Preferences(not impl)",
    NULL,  
    NULL, 
    0, 
    NULL },
*/
  { "/_Help",     
    NULL,  
    NULL,
    0,
    "<LastBranch>" },
  { "/_Help/About",   
    NULL,
    about,
    0, 
    NULL },
};

/**
 * This creates the main window
 **/
static void makeMainWindow() {
  GtkWidget * window;
  GtkWidget * button;
  GtkWidget * hbox;
  GtkWidget * vbox;
  GtkWidget * menubar;
  GtkWidget * table;
  GtkWidget * label;
  GtkAccelGroup * accel_group;
  GtkItemFactory * item_factory;

  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);

  /* create main window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL); 
  gtk_window_set_title(GTK_WINDOW(window), 
		       "GNUnet: gtk+ GUI");
  gtk_widget_set_usize(GTK_WIDGET(window), 
		       780, /* x-size */
		       300); /* y-size */
  vbox = gtk_vbox_new(FALSE, 1);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_signal_connect(GTK_OBJECT(window), 
		      "delete_event",
		      GTK_SIGNAL_FUNC(deleteEvent), 
		     NULL);
  gtk_signal_connect(GTK_OBJECT(window), 
		      "destroy",
		      GTK_SIGNAL_FUNC(destroyMain),
		     NULL);
  gtk_widget_show(vbox);

  
  /* create pulldown menues */
  accel_group = gtk_accel_group_new ();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, 
				      "<main>",
				      accel_group);
  gtk_item_factory_create_items(item_factory, 
				nmenu_items,
				menu_items, 
				NULL);
  gtk_window_add_accel_group(GTK_WINDOW (window), 
			     accel_group);
  menubar = gtk_item_factory_get_widget(item_factory, 
					"<main>");
  gtk_box_pack_start(GTK_BOX (vbox), 
		     menubar,
		     FALSE, 
		     TRUE, 
		     0);
  gtk_widget_show(menubar);


  /* a table to put the search results notebook to */
  table = gtk_table_new(6, 6, TRUE);
  gtk_box_pack_start(GTK_BOX (vbox),
		     table,
		     TRUE,
		     TRUE, 0);
  gtk_widget_show(table);
  
  /* add "notebook" for search results */
  notebook = GTK_NOTEBOOK(gtk_notebook_new());
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook),
			   GTK_POS_TOP);
  gtk_table_attach_defaults(GTK_TABLE(table), 
			    GTK_WIDGET(notebook),
			    0, 6, 0, 6);
  gtk_widget_show(GTK_WIDGET(notebook));

  /* BEGIN of SEARCH BOX CODE */
  /* At the bottom, put the search box */
  hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX (vbox),
		     hbox, 
		     FALSE, 
		     FALSE, 
		     0);
  gtk_widget_show(hbox);
 
  /* search entry label */
  label = gtk_label_new("Keyword(s):");
  gtk_box_pack_start(GTK_BOX(hbox),
		     label,
		     FALSE, 
		     FALSE,
		     0);
  gtk_widget_show(label);

  /* search input line */
  main_window_input_line = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox), 
		     main_window_input_line, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_signal_connect(GTK_OBJECT(main_window_input_line),
		     "activate",
                     GTK_SIGNAL_FUNC(search),
                     notebook);
  gtk_widget_show(main_window_input_line);

  /* search button */
  button = gtk_button_new_with_label("Search");
  gtk_signal_connect(GTK_OBJECT (button), 
		     "clicked",
		     GTK_SIGNAL_FUNC(search), notebook);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  /* END of SEARCH BOX CODE */


  /* show the main window */
  gtk_widget_show(window);
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseOptions(int argc, 
			char ** argv) {
  int shouldexit = OK;
  int c;

  while (1) {
    int option_index=0;
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "debug",   0, 0, 'd' },
      { 0,0,0,0 }
    };
    
    c = GNgetopt_long(argc,
		      argv, 
		      "vhdc:L:H:", 
		      long_options, 
		      &option_index);
    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;    
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-gtk v%s\n",
	     VERSION,
	     AFS_VERSION);
      shouldexit = SYSERR;
      break;
    case 'h': 
      printf("USAGE: gnunet-gtk [OPTIONS]\n");
      printHelpDefault();
      shouldexit = SYSERR;
      break;
    default:
      LOG(LOG_FAILURE,
	  "Unknown option %c. Aborting.\n"
	  "Use --help to get a list of options.\n",
	  c);
      shouldexit = SYSERR;    
    } /* end of parsing commandline */
  }
  if (GNoptind < argc) {
    fprintf(stderr,
	    "Invalid arguments: ");
    while (GNoptind < argc)
      fprintf(stderr,
	      "%s ", 
	      argv[GNoptind++]);
    fprintf(stderr, 
	    "\n");
    shouldexit = SYSERR;
  }
  return shouldexit;
}

/**
 * The main function.
 **/
int main(int argc,
         char * argv[]) {
  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return(0);
  startCron();
  initAnonymityPolicy(NULL);
  refuseToDie = SEMAPHORE_NEW(1);

  g_thread_init(NULL);
  /*  gdk_threads_init(); */
  gtk_init(&argc, &argv);

  makeMainWindow();
  gtk_main();

  LOG(LOG_DEBUG, 
      "DEBUG: GUI leaving...\n");
  SEMAPHORE_FREE(refuseToDie);

  doneAnonymityPolicy();
  stopCron();
  doneUtil();  
  return 0;
}

/* end of main.c */
